"""
This is a collection of transaction-related monitoring utilities.

Usage:

    from edx_django_utils.monitoring import ignore_transaction
    ...
    # called from inside a view you want ignored
    ignore_transaction()

Please remember to expose any new methods in the `__init__.py` file.
"""
try:
    import newrelic.agent
except ImportError:
    newrelic = None  # pylint: disable=invalid-name


def ignore_transaction():
    """
    Ignore the transaction in monitoring. Only works for NewRelic.

    This allows us to ignore code paths that are unhelpful to include, such as
    `/health/` checks.

    """
    # Note: This is not being ported over to backends, because we don't have
    # an equivalent for Datadog. For Datadog, use filter/ignore rules.
    if newrelic:  # pragma: no cover
        newrelic.agent.ignore_transaction()



"""
Signals defined to support monitoring plugins.

Placing signals in this file (vs alternatives like /internal, or
signals/signals.py) provides a good public api for these signals.
"""
import django.dispatch

# The MonitoringSupportMiddleware sends these signals to enable plugging in
# outside monitoring capabilities. This is a useful capability because the
# MonitoringSupportMiddleware should be available in all Open edX services,
# and should be one of the first middleware (after RequestCacheMiddleware),
# allowing it access to monitor most requests (even those that fail or
# respond early in other middleware).
monitoring_support_process_request = django.dispatch.Signal()
monitoring_support_process_response = django.dispatch.Signal()
monitoring_support_process_exception = django.dispatch.Signal()



"""
Tests monitoring middleware.

Note: CachedCustomMonitoringMiddleware is tested in ``test_custom_monitoring.py``.
"""
import re
from unittest.mock import Mock, call, patch

import ddt
from django.core.exceptions import MiddlewareNotUsed
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.test import TestCase
from django.test.client import RequestFactory
from django.test.utils import override_settings
from waffle.testutils import override_switch

from edx_django_utils.cache import RequestCache
from edx_django_utils.monitoring import (
    CookieMonitoringMiddleware,
    DeploymentMonitoringMiddleware,
    FrontendMonitoringMiddleware,
    MonitoringMemoryMiddleware
)


class TestMonitoringMemoryMiddleware(TestCase):
    """
    Tests for MonitoringMemoryMiddleware
    """
    @override_switch('edx_django_utils.monitoring.enable_memory_middleware', False)
    @patch('edx_django_utils.monitoring.internal.middleware.log')
    def test_memory_monitoring_when_disabled(self, mock_logger):
        mock_response = Mock()
        MonitoringMemoryMiddleware(mock_response).process_response(
            'fake request',
            'fake response',
        )
        mock_logger.info.assert_not_called()

    @override_switch('edx_django_utils.monitoring.enable_memory_middleware', True)
    @patch('edx_django_utils.monitoring.internal.middleware.log')
    def test_memory_monitoring_when_enabled(self, mock_logger):
        request = RequestFactory().get('/')
        mock_response = Mock()
        MonitoringMemoryMiddleware(mock_response).process_response(
            request,
            'fake response',
        )
        mock_logger.info.assert_called()


class TestDeploymentMonitoringMiddleware(TestCase):
    """
    Test the DeploymentMonitoringMiddleware functionalities
    """
    # To handle both major.minor and major.minor.patch versions
    version_pattern = r'^\d+\.\d+(\.\d+)?'

    def setUp(self):
        super().setUp()
        RequestCache.clear_all_namespaces()

    def _test_key_value_pair(self, function_call, key):
        """
        Asserts the function call key and value with the provided key and the default version_pattern
        """
        attribute_key, attribute_value = function_call[0]
        assert attribute_key == key
        assert re.match(re.compile(self.version_pattern), attribute_value)

    @patch('newrelic.agent')
    def test_record_python_and_django_version(self, mock_newrelic_agent):
        """
        Test that the DeploymentMonitoringMiddleware records the correct Python and Django versions
        """
        middleware = DeploymentMonitoringMiddleware(Mock())
        middleware(Mock())

        parameter_calls_count = mock_newrelic_agent.add_custom_attribute.call_count
        assert parameter_calls_count == 2

        function_calls = mock_newrelic_agent.add_custom_attribute.call_args_list
        self._test_key_value_pair(function_calls[0], 'python_version')
        self._test_key_value_pair(function_calls[1], 'django_version')


@ddt.ddt
class CookieMonitoringMiddlewareTestCase(TestCase):
    """
    Tests for CookieMonitoringMiddleware.
    """
    def setUp(self):
        super().setUp()
        self.mock_response = Mock()

    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    @patch("edx_django_utils.monitoring.internal.middleware._set_custom_attribute")
    @ddt.data(
        (None, None),  # logging threshold not defined
        (5, None),  # logging threshold too high
        (5, 9999999999999999999),  # logging threshold too high, and random sampling impossibly unlikely
    )
    @ddt.unpack
    def test_cookie_monitoring_with_no_logging(
        self, logging_threshold, sampling_request_count, mock_set_custom_attribute, mock_logger
    ):
        expected_response = self.mock_response
        middleware = CookieMonitoringMiddleware(lambda request: expected_response)
        cookies_dict = {'a': 'y'}

        with override_settings(COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=logging_threshold):
            with override_settings(COOKIE_SAMPLING_REQUEST_COUNT=sampling_request_count):
                actual_response = middleware(self.get_mock_request(cookies_dict))

        assert actual_response == expected_response
        # expect monitoring of header size for all requests
        mock_set_custom_attribute.assert_called_once_with('cookies.header.size', 3)
        # cookie logging was not enabled, so nothing should be logged
        mock_logger.info.assert_not_called()
        mock_logger.exception.assert_not_called()

    @override_settings(COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=None)
    @override_settings(COOKIE_SAMPLING_REQUEST_COUNT=None)
    @patch("edx_django_utils.monitoring.internal.middleware._set_custom_attribute")
    @ddt.data(
        # A corrupt cookie header contains "Cookie: ".
        ('corruptCookie: normal-cookie=value', 1, 1),
        ('corrupt1Cookie: normal-cookie1=value1;corrupt2Cookie: normal-cookie2=value2', 2, 2),
        ('corrupt=Cookie: value', 1, 0),
    )
    @ddt.unpack
    def test_cookie_header_corrupt_monitoring(
        self, corrupt_cookie_header, expected_corrupt_count, expected_corrupt_key_count, mock_set_custom_attribute
    ):
        middleware = CookieMonitoringMiddleware(self.mock_response)
        request = RequestFactory().request()
        request.META['HTTP_COOKIE'] = corrupt_cookie_header

        middleware(request)

        mock_set_custom_attribute.assert_has_calls([
            call('cookies.header.size', len(request.headers['cookie'])),
            call('cookies.header.corrupt_count', expected_corrupt_count),
            call('cookies.header.corrupt_key_count', expected_corrupt_key_count),
        ])

    @override_settings(COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=1)
    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    @patch("edx_django_utils.monitoring.internal.middleware._set_custom_attribute")
    def test_log_cookie_with_threshold_met(self, mock_set_custom_attribute, mock_logger):
        middleware = CookieMonitoringMiddleware(self.mock_response)
        cookies_dict = {
            "a": "yy",
            "b": "xxx",
            "c": "z",
        }

        middleware(self.get_mock_request(cookies_dict))

        mock_set_custom_attribute.assert_has_calls([
            call('cookies.header.size', 16),
            call('cookies.header.size.computed', 16)
        ])
        mock_logger.info.assert_called_once_with(
            "Large (>= 1) cookie header detected. BEGIN-COOKIE-SIZES(total=16) b: 3, a: 2, c: 1 END-COOKIE-SIZES"
        )
        mock_logger.exception.assert_not_called()

    @override_settings(
        COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=1,
        UNUSUAL_COOKIE_HEADER_PUBLIC_KEY="some private key",
    )
    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    @patch("edx_django_utils.monitoring.internal.middleware.encrypt_for_log", return_value="[encrypted: 50M3JUN|<]")
    def test_log_corrupt_cookies(self, mock_encrypt, mock_logger):
        middleware = CookieMonitoringMiddleware(self.mock_response)
        request = RequestFactory().request()
        request.META['HTTP_COOKIE'] = 'aa=1; bCookie: bb=22; ccc=3Cookie: 33'
        request.META['HTTP_SOMETHING'] = 'else'

        middleware(request)

        # Is passed all headers
        mock_encrypt.assert_called_once_with(
            '{"Cookie": "aa=1; bCookie: bb=22; ccc=3Cookie: 33", "Something": "else"}',
            "some private key",
        )

        # Encrypted headers are logged in additional message if any cookies appeared corrupted.
        mock_logger.info.assert_has_calls([
            call("All headers for request with corrupted cookies (count=2): [encrypted: 50M3JUN|<]"),
            call(
                "Large (>= 1) cookie header detected. "
                "BEGIN-COOKIE-SIZES(total=37) ccc: 11, bCookie: bb: 2, aa: 1 END-COOKIE-SIZES"
            ),
        ])

    @override_settings(
        COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=1,
        UNUSUAL_COOKIE_HEADER_PUBLIC_KEY="some private key",
        UNUSUAL_COOKIE_HEADER_LOG_CHUNK=75,
    )
    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    @patch(
        "edx_django_utils.monitoring.internal.middleware.encrypt_for_log",
        return_value="[encrypted: 50M3JUN|<_aaaabbbbccccdddd]"
    )
    def test_log_corrupt_cookies_split(self, mock_encrypt, mock_logger):
        """
        Like ``test_log_corrupt_cookies`` but with message splitting.
        """
        middleware = CookieMonitoringMiddleware(self.mock_response)
        request = RequestFactory().request()
        request.META['HTTP_COOKIE'] = 'aa=1; bCookie: bb=22; ccc=3Cookie: 33'
        request.META['HTTP_SOMETHING'] = 'else'

        middleware(request)

        # Is passed all headers
        mock_encrypt.assert_called_once_with(
            '{"Cookie": "aa=1; bCookie: bb=22; ccc=3Cookie: 33", "Something": "else"}',
            "some private key",
        )

        # Encrypted headers are logged across multiple messages if too large.
        mock_logger.info.assert_has_calls([
            call(
                "All headers for request with corrupted cookies (count=2): "
                "[encrypted: 50M3J [chunk #1, group=OUMTkx5O continues]"
            ),
            call("UN|<_aaaabbbbccccdddd] [chunk #2, group=OUMTkx5O final]"),
            call(
                "Large (>= 1) cookie header detected. "
                "BEGIN-COOKIE-SIZES(total=37) ccc: 11, bCookie: bb: 2, aa: 1 END-COOKIE-SIZES"
            ),
        ])

    @override_settings(COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=9999)
    @override_settings(COOKIE_SAMPLING_REQUEST_COUNT=1)
    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    @patch("edx_django_utils.monitoring.internal.middleware._set_custom_attribute")
    def test_log_cookie_with_sampling(self, mock_set_custom_attribute, mock_logger):
        middleware = CookieMonitoringMiddleware(self.mock_response)
        cookies_dict = {
            "a": "yy",
            "b": "xxx",
            "c": "z",
        }

        middleware(self.get_mock_request(cookies_dict))

        mock_set_custom_attribute.assert_has_calls([
            call('cookies.header.size', 16),
            call('cookies.header.size.computed', 16)
        ])
        mock_logger.info.assert_called_once_with(
            "Sampled small (< 9999) cookie header. BEGIN-COOKIE-SIZES(total=16) b: 3, a: 2, c: 1 END-COOKIE-SIZES"
        )
        mock_logger.exception.assert_not_called()

    @override_settings(COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=9999)
    @override_settings(COOKIE_SAMPLING_REQUEST_COUNT=1)
    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    @patch("edx_django_utils.monitoring.internal.middleware._set_custom_attribute")
    def test_empty_cookie_header_skips_sampling(self, mock_set_custom_attribute, mock_logger):
        middleware = CookieMonitoringMiddleware(self.mock_response)
        cookies_dict = {}

        middleware(self.get_mock_request(cookies_dict))

        mock_set_custom_attribute.assert_has_calls([
            call('cookies.header.size', 0),
        ])
        mock_logger.info.assert_not_called()
        mock_logger.exception.assert_not_called()

    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    def test_cookie_monitoring_unknown_exception(self, mock_logger):
        middleware = CookieMonitoringMiddleware(self.mock_response)
        cookies_dict = {'a': 'y'}
        mock_request = self.get_mock_request(cookies_dict)
        mock_request.META = Mock()
        mock_request.META.side_effect = Exception("Some exception")

        middleware(mock_request)

        mock_logger.exception.assert_called_once_with("Unexpected error logging and monitoring cookies.")

    @override_settings(COOKIE_PREFIXES_TO_REMOVE=[('old_cookie', 'localhost')])
    def test_deprecated_cookies_removed(self):
        self.mock_response.reset_mock()
        middleware = CookieMonitoringMiddleware(lambda _: self.mock_response)
        cookies_dict = {'old_cookie': 'x',
                        'ok_cookie': 'x',
                        'old_cookie_9000': 'x'
                        }
        response = middleware(self.get_mock_request(cookies_dict))

        assert response == self.mock_response
        assert response.delete_cookie.mock_calls == [
            call('old_cookie',  domain='localhost'),
            call('old_cookie_9000', domain='localhost'),
        ]

    @override_settings(COOKIE_PREFIXES_TO_REMOVE='old_cookie')
    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    def test_bad_cookie_prefix_setting(self, mock_log):
        self.mock_response.reset_mock()
        middleware = CookieMonitoringMiddleware(lambda _: self.mock_response)
        cookies_dict = {'old_cookie': 'x',
                        'ok_cookie': 'x',
                        'old_cookie_9000': 'x'
                        }
        response = middleware(self.get_mock_request(cookies_dict))

        assert response == self.mock_response
        mock_log.warning.assert_called_once_with("COOKIE_PREFIXES_TO_REMOVE must be a list of (name, domain) tuples,"
                                                 " not <class 'str'>. No cookies will be removed.")
        response.delete_cookie.assert_not_called()

    def get_mock_request(self, cookies_dict):
        """
        Return mock request with the provided cookies in the header.
        """
        factory = RequestFactory()
        for name, value in cookies_dict.items():
            factory.cookies[name] = value
        return factory.request()


@ddt.ddt
class FrontendMonitoringMiddlewareTestCase(TestCase):
    """
    Tests for FrontendMonitoringMiddleware.
    """
    def setUp(self):
        super().setUp()
        self.script = "<script>test script</script>"

    @override_switch('edx_django_utils.monitoring.enable_frontend_monitoring_middleware', False)
    def test_frontend_middleware_with_waffle_diasbled(self):
        """
        Test that middleware is disabled when waffle flag is not enabled.
        """
        original_html = '<head></head>'
        with override_settings(OPENEDX_TELEMETRY_FRONTEND_SCRIPTS=self.script):
            self.assertRaises(
                MiddlewareNotUsed,
                FrontendMonitoringMiddleware,
                lambda r: HttpResponse(original_html, content_type='text/html')
            )

    @override_switch('edx_django_utils.monitoring.enable_frontend_monitoring_middleware', True)
    def test_frontend_middleware_with_waffle_enabled(self):
        """
        Test that middleware works as expected when flag is enabled.
        """
        with override_settings(OPENEDX_TELEMETRY_FRONTEND_SCRIPTS=self.script):
            middleware = FrontendMonitoringMiddleware(lambda r: HttpResponse('<head></head>', content_type='text/html'))
            response = middleware(HttpRequest())
        # Assert that the script is inserted into the response when flag is enabled
        assert self.script.encode() in response.content

    @override_switch('edx_django_utils.monitoring.enable_frontend_monitoring_middleware', True)
    @patch("edx_django_utils.monitoring.internal.middleware.FrontendMonitoringMiddleware.inject_script")
    def test_frontend_middleware_without_setting_variable(self, mock_inject_script):
        """
        Test that middleware behaves correctly when setting variable is not defined.
        """
        original_html = '<html><head></head><body></body><html>'
        middleware = FrontendMonitoringMiddleware(lambda r: HttpResponse(original_html, content_type='text/html'))
        response = middleware(HttpRequest())
        # Assert that the response content remains unchanged if settings not defined
        assert response.content == original_html.encode()

        mock_inject_script.assert_not_called()

    @override_switch('edx_django_utils.monitoring.enable_frontend_monitoring_middleware', True)
    @patch("edx_django_utils.monitoring.internal.middleware.FrontendMonitoringMiddleware.inject_script")
    def test_frontend_middleware_for_json_requests(self, mock_inject_script):
        """
        Test that middleware doesn't insert script tag for json requests
        """
        middleware = FrontendMonitoringMiddleware(lambda r: JsonResponse({"dummy": True}))
        response = middleware(HttpRequest())
        # Assert that the response content remains unchanged if settings not defined
        assert response.content == b'{"dummy": true}'

        mock_inject_script.assert_not_called()

    @override_switch('edx_django_utils.monitoring.enable_frontend_monitoring_middleware', True)
    @ddt.data(
        ('<html><body></body><html>', '<body>'),
        ('<html><head></head><body></body><html>', '</head>'),
        ('<head></head><body></body>', '</head>'),
        ('<body></body>', '<body>'),
        ('<head></head>', '</head>'),
    )
    @ddt.unpack
    def test_frontend_middleware_with_head_and_body_tag(self, original_html, expected_tag):
        """
        Test that script is inserted at the right place.
        """
        with override_settings(OPENEDX_TELEMETRY_FRONTEND_SCRIPTS=self.script):
            middleware = FrontendMonitoringMiddleware(lambda r: HttpResponse(original_html, content_type='text/html'))
            response = middleware(HttpRequest())

        # Assert that the script is inserted at the right place
        assert f"{self.script}{expected_tag}".encode() in response.content

    @override_switch('edx_django_utils.monitoring.enable_frontend_monitoring_middleware', True)
    @ddt.data(
        '<html></html>',
        '<center></center>',
    )
    def test_frontend_middleware_without_head_and_body_tag(self, original_html):
        """
        Test that middleware behavior is correct when both of head and body tag are missing in the response.
        """
        with override_settings(OPENEDX_TELEMETRY_FRONTEND_SCRIPTS=self.script):
            middleware = FrontendMonitoringMiddleware(lambda r: HttpResponse(original_html, content_type='text/html'))
            response = middleware(HttpRequest())
        # Assert that the response content remains unchanged if no body tag is found
        assert response.content == original_html.encode()

    @override_switch('edx_django_utils.monitoring.enable_frontend_monitoring_middleware', True)
    def test_frontend_middleware_content_length_header_already_set(self):
        """
        Test that middleware updates the Content-Length header, when its already set.
        """
        original_html = '<head></head>'
        with override_settings(OPENEDX_TELEMETRY_FRONTEND_SCRIPTS=self.script):
            middleware = FrontendMonitoringMiddleware(lambda r: HttpResponse(
                original_html, content_type='text/html', headers={'Content-Length': len(original_html)}))
            response = middleware(HttpRequest())
        # Assert that the response content contains script tag
        assert self.script.encode() in response.content
        # Assert that the Content-Length header is updated and script length is added.
        assert response.headers.get('Content-Length') == str(len(original_html) + len(self.script))

    @override_switch('edx_django_utils.monitoring.enable_frontend_monitoring_middleware', True)
    def test_frontend_middleware_content_length_header_not_set(self):
        """
        Test that middleware doesn't set the Content-Length header when it's not already set.
        """
        original_html = '<head></head>'
        with override_settings(OPENEDX_TELEMETRY_FRONTEND_SCRIPTS=self.script):
            middleware = FrontendMonitoringMiddleware(lambda r: HttpResponse(original_html, content_type='text/html'))
            response = middleware(HttpRequest())
        # Assert that the response content contains script tag
        assert self.script.encode() in response.content
        # Assert that the Content-Length header isn't updated, when not set already
        assert response.headers.get('Content-Length') is None



"""
Test management commands and related functions.
"""

import mock
from django.core.management import get_commands, load_command_class
from django.core.management.base import CommandError, SystemCheckError
from django.test import TestCase
from pytest import mark

from edx_django_utils.data_generation.management.commands.manufacture_data import Command, Node
# We need to import factories even if we don't use them directly, in order for them to be picked up
# by the manufacture_data command's discovery process
# pylint: disable=unused-import
from edx_django_utils.data_generation.tests.factories import (
    TestCompanyFactory,
    TestContactInfoFactory,
    TestModelNonstandardCasingFactory,
    TestNationalIdFactory,
    TestPersonContactPhoneNumberFactory,
    TestPersonFactory
)
from edx_django_utils.data_generation.tests.models import (
    TestCompany,
    TestContactInfo,
    TestNationalId,
    TestPerson,
    TestPersonContactPhoneNumber,
    test_model_nonstandard_casing
)


class TestCommand(Command):
    """
    Class for use in testing manufacture_data command via run_from_argv
    """

    def check(self, *args):
        # Skip checks that aren't needed or configured in test suite
        pass


# Copied from django.core.management.__init__.py, with arg checking disabled given the open-ended nature of the
# model customizations we might need to specify.
# https://github.com/django/django/blob/1ad7761ee616341295f36c80f78b86ff79d5b513/django/core/management/__init__.py#L83
def call_command(command_name, *args, **options):
    """
    Call the given command, with the given options and args/kwargs.

    This is the primary API you should use for calling specific commands.

    `command_name` may be a string or a command object. Using a string is
    preferred unless the command object is required for further processing or
    testing.

    Some examples:
        call_command('migrate')
        call_command('shell', plain=True)
        call_command('sqlmigrate', 'myapp')

        from django.core.management.commands import flush
        cmd = flush.Command()
        call_command(cmd, verbosity=0, interactive=False)
        # Do something with cmd ...
    """
    app_name = get_commands()[command_name]
    command = load_command_class(app_name, command_name)

    # Simulate argument parsing to get the option defaults (see #10080 for details).
    parser = command.create_parser("", command_name)
    # Use the `dest` option name from the parser option
    opt_mapping = {
        min(s_opt.option_strings).lstrip("-").replace("-", "_"): s_opt.dest
        for s_opt in parser._actions  # pylint: disable=protected-access
        if s_opt.option_strings
    }
    arg_options = {opt_mapping.get(key, key): value for key, value in options.items()}
    parse_args = []

    defaults = parser.parse_args(args=parse_args)

    # pylint: disable=protected-access
    defaults = dict(
        defaults._get_kwargs(), **arg_options
    )

    # Move positional args out of options to mimic legacy optparse
    args = defaults.pop("args", ())
    if "skip_checks" not in options:
        defaults["skip_checks"] = True

    return command.execute(*args, **defaults)


@mark.django_db
class ManufactureDataCommandTests(TestCase):
    """
    Test command `manufacture_data`.
    """

    command = "manufacture_data"

    def test_command_requires_model(self):
        """
        Test that the manufacture_data command will raise an error if no model is provided.
        """
        with self.assertRaises(CommandError):
            call_command(self.command)

    def test_command_requires_valid_model(self):
        """
        Test that the manufacture_data command will raise an error if the provided model is invalid.
        """
        with self.assertRaises(CommandError):
            call_command(self.command, model="FakeModel")

    def test_single_object_create_no_customizations(self):
        """
        Test that the manufacture_data command will create a single object with no customizations.
        """
        assert TestPerson.objects.all().count() == 0
        created_object = call_command(
            self.command,
            model="edx_django_utils.data_generation.tests.models.TestPerson",
        )
        assert TestPerson.objects.all().count() == 1
        assert TestPerson.objects.filter(pk=created_object).exists()

    def test_command_requires_valid_field(self):
        """
        Test that the manufacture_data command will raise an error if the provided field is invalid.
        """
        with self.assertRaises(CommandError):
            call_command(
                self.command,
                model="TestPerson",
                field_customizations={"fake_field": "fake_value"},
            )

    def test_command_can_customize_fields(self):
        """
        Test that the manufacture_data command will create a single object with customizations.
        """
        assert TestPerson.objects.all().count() == 0
        created_object = call_command(
            self.command,
            model="edx_django_utils.data_generation.tests.models.TestPerson",
            field_customizations={"first_name": "Steve"},
        )
        assert TestPerson.objects.all().count() == 1
        assert TestPerson.objects.filter(pk=created_object).exists()
        assert (
            TestPerson.objects.filter(pk=created_object).first().first_name == "Steve"
        )

    def test_command_can_customize_nested_objects(self):
        """
        Test that the manufacture_data command supports customizing nested objects.
        """
        assert TestPerson.objects.all().count() == 0
        assert TestContactInfo.objects.all().count() == 0
        created_object = call_command(
            self.command,
            model="edx_django_utils.data_generation.tests.models.TestContactInfo",
            field_customizations={
                "address": "123 4th st",
                "test_person__first_name": "Joey",
                "test_person__last_name": "Nowhere",
            },
        )
        assert TestPerson.objects.all().count() == 2
        assert TestContactInfo.objects.all().count() == 1
        assert (
            TestContactInfo.objects.filter(pk=created_object)
            .first()
            .test_person.last_name
            == "Nowhere"
        )

    def test_command_can_customize_doubly_nested_objects(self):
        """
        Test that the manufacture_data command supports customizing nested objects.
        """
        assert TestPerson.objects.all().count() == 0
        assert TestContactInfo.objects.all().count() == 0
        created_object = call_command(
            self.command,
            model="edx_django_utils.data_generation.tests.models.TestPersonContactPhoneNumber",
            field_customizations={
                "phone_number": "(000) 000-0000",
                "test_contact_info__test_person__first_name": "Joey",
                "test_contact_info__test_person__last_name": "Nowhere",
                "test_contact_info__address": "123 4th st",
            },
        )
        assert TestPerson.objects.all().count() == 2
        assert TestContactInfo.objects.all().count() == 1
        assert (
            TestContactInfo.objects.filter(pk=created_object)
            .first()
            .test_person.last_name
            == "Nowhere"
        )

    def test_command_can_customize_multiple_nested_objects(self):
        """
        Test that the manufacture_data command supports customizing nested objects.
        """
        assert TestPerson.objects.all().count() == 0
        assert TestContactInfo.objects.all().count() == 0
        created_object = call_command(
            self.command,
            model="edx_django_utils.data_generation.tests.models.TestPersonContactPhoneNumber",
            field_customizations={
                "phone_number": "(000) 000-0000",
                "test_contact_info__test_person__first_name": "Joey",
                "test_contact_info__test_company__company_name": "JoeyCo",
                "test_contact_info__address": "123 4th st",
            },
        )
        assert TestPerson.objects.all().count() == 2
        assert TestContactInfo.objects.all().count() == 1
        assert (
            TestContactInfo.objects.filter(pk=created_object)
            .first()
            .test_person.first_name
            == "Joey"
        )
        assert (
            TestCompany.objects.filter(company_name='JoeyCo')
            .count() == 1
        )

    def test_command_can_customize_nested_objects_with_fk(self):
        """
        Test that the manufacture_data command supports customizing nested objects.
        """
        assert TestPerson.objects.all().count() == 0
        assert TestContactInfo.objects.all().count() == 0
        test_id = call_command(
            self.command,
            model="edx_django_utils.data_generation.tests.models.TestNationalId",
        )

        call_command(
            self.command,
            model="edx_django_utils.data_generation.tests.models.TestPersonContactPhoneNumber",
            field_customizations={
                "test_contact_info__test_person__national_id": test_id,
                "test_contact_info__test_company__national_id": test_id,
            },
        )
        assert TestPerson.objects.all().count() == 2
        assert TestCompany.objects.all().count() == 1
        assert TestContactInfo.objects.all().count() == 1
        assert TestNationalId.objects.filter(pk=test_id).count() == 1

    def test_command_builds_chains_of_pk(self):
        """
        Test that the manufacture_data command supports customizing nested objects.
        """
        assert TestPerson.objects.all().count() == 0
        assert TestContactInfo.objects.all().count() == 0
        joe = TestPersonFactory()

        call_command(
            self.command,
            model="edx_django_utils.data_generation.tests.models.TestPersonContactPhoneNumber",
            field_customizations={
                "test_contact_info__test_person": f"{joe.pk}",
                "test_contact_info__test_company__ceo": f"{joe.pk}",
            },
        )
        assert TestPerson.objects.all().count() == 1
        assert TestCompany.objects.all().count() == 1
        assert TestContactInfo.objects.all().count() == 1

    def test_command_cannot_edit_created_fk(self):
        """
        Error case: trying to edit foreign key after foreign key object has already been created
        """
        assert TestPerson.objects.all().count() == 0
        assert TestContactInfo.objects.all().count() == 0

        with self.assertRaises(CommandError) as cm:
            call_command(
                self.command,
                model="edx_django_utils.data_generation.tests.models.TestPersonContactPhoneNumber",
                field_customizations={
                    "phone_number": "(000) 000-0000",
                    "test_contact_info__test_person__first_name": "Joey",
                    "test_contact_info__test_person": "0",
                },
            )
        assert str(cm.exception) == 'Provided FK value: 0 does not exist on TestPerson'

    def test_command_cannot_customize_foreign_keys(self):
        """
        Error case: customizing provided objects.
        """
        assert TestPerson.objects.all().count() == 0
        assert TestContactInfo.objects.all().count() == 0
        test_person = call_command(
            self.command,
            model="edx_django_utils.data_generation.tests.models.TestPerson",
            field_customizations={"first_name": "Steve"},
        )
        with self.assertRaises(CommandError) as cm:
            call_command(
                self.command,
                model="edx_django_utils.data_generation.tests.models.TestContactInfo",
                field_customizations={
                    "address": "123 4th st",
                    "test_person": test_person,
                    "test_person__last_name": "Harvey",
                },
            )
        assert str(cm.exception) == 'This script cannot modify existing objects'

    def test_command_cannot_customize_nested_foreign_keys(self):
        """
        Error case: customizing nested provided objects.
        """
        assert TestPerson.objects.all().count() == 0
        assert TestContactInfo.objects.all().count() == 0
        test_person = call_command(
            self.command,
            model="edx_django_utils.data_generation.tests.models.TestPerson",
            field_customizations={"first_name": "Steve"},
        )
        with self.assertRaises(CommandError) as cm:
            call_command(
                self.command,
                model="edx_django_utils.data_generation.tests.models.TestPersonContactPhoneNumber",
                field_customizations={
                    "phone_number": "(000) 000-0000",
                    "test_contact_info__test_person__first_name": "Joey",
                    "test_contact_info__test_person": test_person,
                    "test_contact_info__address": "123 4th st",
                },
            )
        assert str(cm.exception) == 'This script does not support customizing provided existing objects'

    def test_command_object_foreign_key(self):
        """
        Test that the manufacture_data command supports creating objects with foreign keys
        """
        assert TestPerson.objects.all().count() == 0
        foreign_key_object_id = call_command(
            self.command,
            model="edx_django_utils.data_generation.tests.models.TestPerson",
            field_customizations={"first_name": "Steve"},
        )
        assert TestPerson.objects.all().count() == 1
        created_object = call_command(
            self.command,
            model="edx_django_utils.data_generation.tests.models.TestContactInfo",
            field_customizations={"test_person": foreign_key_object_id},
        )
        assert (
            TestContactInfo.objects.filter(pk=created_object)
            .first()
            .test_person.first_name
            == "Steve"
        )

    def test_argv_command_can_customize_nested_objects(self):
        """
        argv: Test that the manufacture_data command supports customizing nested objects.
        """
        assert TestPerson.objects.all().count() == 0
        assert TestContactInfo.objects.all().count() == 0
        command = TestCommand()

        command.run_from_argv(
            [
                "manage.py",
                "manufacture_data",
                "--model",
                "edx_django_utils.data_generation.tests.models.TestContactInfo",
                "--test_person__last_name",
                "Nowhere",
            ]
        )
        assert TestPerson.objects.all().count() == 2
        assert TestContactInfo.objects.all().count() == 1
        assert TestContactInfo.objects.first().test_person.last_name == "Nowhere"

    def test_argv_command_error(self):
        """
        argv error: Nested model does not exist
        """
        assert TestPerson.objects.all().count() == 0
        assert TestContactInfo.objects.all().count() == 0
        command = TestCommand()

        with self.assertRaises(SystemExit):
            command.run_from_argv(
                [
                    "manage.py",
                    "manufacture_data",
                    "--model",
                    "edx_django_utils.data_generation.tests.models.ThisModelDoesNotExist",
                ]
            )

    @mock.patch('edx_django_utils.data_generation.management.commands.manufacture_data.Command.handle')
    def test_argv_system_check_error(self, handleMock):
        """
        argv error: SystemCheckError
        """
        handleMock.side_effect = SystemCheckError(mock.Mock('SystemCheckError'))
        command = TestCommand()

        with self.assertRaises(SystemExit):
            command.run_from_argv(
                [
                    "manage.py",
                    "manufacture_data",
                    "--model",
                    "edx_django_utils.data_generation.tests.models.TestPerson",
                ]
            )

    def test_nonstandard_casing(self):
        """
        Test that the manufacture_data command will work with models that use non-standard casing
        """
        assert test_model_nonstandard_casing.objects.all().count() == 0
        created_object = call_command(
            self.command,
            model="edx_django_utils.data_generation.tests.models.test_model_nonstandard_casing",
        )
        assert test_model_nonstandard_casing.objects.all().count() == 1
        assert test_model_nonstandard_casing.objects.filter(pk=created_object).exists()

    def test_command_nested_nonexistent_model(self):
        """
        Error case: Nested model does not exist
        """
        with self.assertRaises(CommandError):
            call_command(
                self.command,
                model="edx_django_utils.data_generation.tests.models.TestContactInfo",
                field_customizations={
                    "address": "123 4th st",
                    "test_nonperson__last_name": "non-name",
                },
            )

    def test_command_nested_nonexistent_attribute(self):
        """
        Error case: Nested model does not exist
        """
        with self.assertRaises(CommandError):
            call_command(
                self.command,
                model="edx_django_utils.data_generation.tests.models.TestContactInfo",
                field_customizations={
                    "address": "123 4th st",
                    "test_person__middle_name": "Milhaus",
                },
            )

    def test_node_no_factory(self):
        """
        Node error case: no factory provided
        """
        node = Node(field_name='field', model_name='model')
        with self.assertRaises(CommandError):
            node.build_records()

    def test_node_str(self):
        """
        Node __str__ test
        """
        node = Node(field_name='model', model_name='model')
        node.add_child(Node(field_name='field', field_path='model.field', model_name='model'))
        assert str(node) == "<Tree Node (model)> fields: {}\n\t<Tree Node model.field(model)> fields: {}\n"



"""
Tests for TelemetryBackend and implementations.
"""
from unittest.mock import patch

import ddt
import pytest
from django.test import TestCase, override_settings

from edx_django_utils.monitoring import record_exception, set_custom_attribute
from edx_django_utils.monitoring.internal.backends import configured_backends


@ddt.ddt
class TestBackendsConfig(TestCase):
    """
    Test configuration of the backends list.
    """

    def _get_configured_classnames(self):
        return [b.__class__.__name__ for b in configured_backends()]

    @ddt.data(
        # Default
        (None, ['NewRelicBackend']),
        # Empty list removes all
        ([], []),
        # New Relic not required
        (
            ['edx_django_utils.monitoring.OpenTelemetryBackend'],
            ['OpenTelemetryBackend']
        ),
        # All known classes
        (
            [
                'edx_django_utils.monitoring.NewRelicBackend',
                'edx_django_utils.monitoring.OpenTelemetryBackend',
                'edx_django_utils.monitoring.DatadogBackend',
            ],
            ['NewRelicBackend', 'OpenTelemetryBackend', 'DatadogBackend'],
        ),
    )
    @ddt.unpack
    def test_configured_backends(self, setting, expected_classnames):
        """
        Test that backends are loaded as expected.
        """
        with override_settings(OPENEDX_TELEMETRY=setting):
            backend_classnames = self._get_configured_classnames()
            assert sorted(backend_classnames) == sorted(expected_classnames)

    def test_type(self):
        """
        Test that we detect the misuse of a string instead of a list.
        """
        with override_settings(OPENEDX_TELEMETRY='edx_django_utils.monitoring.NewRelicBackend'):
            with pytest.raises(Exception, match='must be a list, not a string'):
                self._get_configured_classnames()

    @patch('edx_django_utils.monitoring.internal.backends.log')
    def test_import_failure(self, mock_log):
        """
        Test that backends that can't be imported are ignored, with warning.
        """
        with override_settings(OPENEDX_TELEMETRY=[
                'nonsense',
                'edx_django_utils.monitoring.OpenTelemetryBackend',
        ]):
            assert self._get_configured_classnames() == ['OpenTelemetryBackend']
        mock_log.warning.assert_called_once_with(
            "Could not load OPENEDX_TELEMETRY option 'nonsense': "
            """ImportError("nonsense doesn't look like a module path")"""
        )

    @patch('edx_django_utils.monitoring.internal.backends.log')
    def test_wrong_class(self, mock_log):
        """
        Test that backend classes of an unexpected ancestor are ignored, with warning.
        """
        with override_settings(OPENEDX_TELEMETRY=[
                'builtins.dict',
                'edx_django_utils.monitoring.OpenTelemetryBackend',
        ]):
            assert self._get_configured_classnames() == ['OpenTelemetryBackend']
        mock_log.warning.assert_called_once_with(
            "Could not load OPENEDX_TELEMETRY option 'builtins.dict': "
            "<class 'dict'> is not a subclass of TelemetryBackend"
        )

    @patch('edx_django_utils.monitoring.internal.backends.log')
    @patch('edx_django_utils.monitoring.internal.backends.newrelic', None)
    def test_newrelic_package(self, mock_log):
        """
        Test that New Relic backend is skipped if package not present.
        """
        with override_settings(OPENEDX_TELEMETRY=['edx_django_utils.monitoring.NewRelicBackend']):
            assert self._get_configured_classnames() == []
        mock_log.warning.assert_called_once_with(
            "Could not load OPENEDX_TELEMETRY option 'edx_django_utils.monitoring.NewRelicBackend': "
            "Exception('Could not load New Relic monitoring backend; package not present.')"
        )

    def test_default_config(self):
        """
        We need to keep the same unconfigured default for now.
        """
        assert [b.__class__.__name__ for b in configured_backends()] == ['NewRelicBackend']


class TestBackendsFanOut(TestCase):
    """
    Test that certain utility functions fan out to the backends.
    """

    @patch('newrelic.agent.add_custom_attribute')
    @patch('opentelemetry.trace.span.NonRecordingSpan.set_attribute')
    # Patch out the span-getter, not the set_attribute call, because
    # it doesn't give us a span unless one is active. And I didn't
    # feel like setting that up.
    #
    # This does at least assert that we're getting the *root* span for DD.
    @patch('ddtrace._trace.tracer.Tracer.current_root_span')
    def test_set_custom_attribute(
            self, mock_dd_root_span,
            mock_otel_set_attribute, mock_nr_add_custom_attribute,
    ):
        with override_settings(OPENEDX_TELEMETRY=[
                'edx_django_utils.monitoring.NewRelicBackend',
                'edx_django_utils.monitoring.OpenTelemetryBackend',
                'edx_django_utils.monitoring.DatadogBackend',
        ]):
            set_custom_attribute('some_key', 'some_value')
        mock_nr_add_custom_attribute.assert_called_once_with('some_key', 'some_value')
        mock_otel_set_attribute.assert_called_once()
        mock_dd_root_span.assert_called_once()

    @patch('newrelic.agent.notice_error')
    @patch('opentelemetry.trace.span.NonRecordingSpan.record_exception')
    # Record exception on current span, not root span.
    @patch('ddtrace._trace.tracer.Tracer.current_span')
    def test_record_exception(
            self, mock_dd_span,
            mock_otel_record_exception, mock_nr_notice_error,
    ):
        with override_settings(OPENEDX_TELEMETRY=[
                'edx_django_utils.monitoring.NewRelicBackend',
                'edx_django_utils.monitoring.OpenTelemetryBackend',
                'edx_django_utils.monitoring.DatadogBackend',
        ]):
            record_exception()
        mock_nr_notice_error.assert_called_once()
        mock_otel_record_exception.assert_called_once()
        mock_dd_span.assert_called_once()



"""
Telemetry abstraction and backends that implement it.

Only a certain subset of the monitoring functions have been made
configurable via this module.
"""

import logging
import sys
from abc import ABC, abstractmethod
from functools import lru_cache

from django.conf import settings
from django.dispatch import receiver
from django.test.signals import setting_changed
from django.utils.module_loading import import_string

log = logging.getLogger(__name__)

# The newrelic package used to not be part of the requirements files
# and so a try-import was used here. This situation is no longer true,
# but we're still preserving that pattern until someone feels like
# doing the work to remove it. (Should just be a major version bump
# and communication to anyone who might be specifically removing the
# package for some reason.)
#
# Ticket for just doing an unconditional import:
# https://github.com/openedx/edx-django-utils/issues/396
try:
    import newrelic.agent
except ImportError:  # pragma: no cover
    newrelic = None  # pylint: disable=invalid-name


class TelemetryBackend(ABC):
    """
    Base class for telemetry sinks.
    """
    @abstractmethod
    def set_attribute(self, key, value):
        """
        Set a key-value attribute on a span. This might be the current
        span or it might the root span of the process, depending on
        the backend.
        """

    @abstractmethod
    def record_exception(self):
        """
        Record the exception that is currently being handled.
        """

    @abstractmethod
    def create_span(self, name):
        """
        Start a tracing span with the given name, returning a context manager instance.

        The caller must use the return value in a `with` statement or similar so that the
        span is guaranteed to be closed appropriately.

        Implementations should create a new child span parented to the current span,
        or create a new root span if not currently in a span.
        """

    @abstractmethod
    def tag_root_span_with_error(self, exception):
        """
        Tags the root span with the given exception. This is primarily useful for
        Datadog as New Relic handles this behavior correctly. Unclear if this is also needs to
        be implemented for OTEL.
        """

    @abstractmethod
    def set_local_root_span_name(self, name, group=None, priority=None):
        """
        Sets the name, group, and priority for a span.
        """


class NewRelicBackend(TelemetryBackend):
    """
    Send telemetry to New Relic.

    API reference: https://docs.newrelic.com/docs/apm/agents/python-agent/python-agent-api/guide-using-python-agent-api/
    """
    def __init__(self):
        if newrelic is None:
            raise Exception("Could not load New Relic monitoring backend; package not present.")

    def set_attribute(self, key, value):
        # Sets attribute on the transaction, rather than the current
        # span, matching historical behavior. There is also an
        # `add_custom_span_attribute` that would better match
        # OpenTelemetry's behavior, which we could try exposing
        # through a new, more specific TelemetryBackend method.
        newrelic.agent.add_custom_attribute(key, value)

    def record_exception(self):
        newrelic.agent.notice_error()

    def create_span(self, name):
        if newrelic.version_info[0] >= 5:
            return newrelic.agent.FunctionTrace(name)
        else:
            nr_transaction = newrelic.agent.current_transaction()
            return newrelic.agent.FunctionTrace(nr_transaction, name)

    def tag_root_span_with_error(self, exception):
        # Does not need to be implemented for NewRelic, because it is handled automatically.
        pass

    def set_local_root_span_name(self, name, group=None, priority=None):
        newrelic.agent.set_transaction_name(name, group, priority)


class OpenTelemetryBackend(TelemetryBackend):
    """
    Send telemetry via OpenTelemetry.

    API reference: https://opentelemetry-python.readthedocs.io/en/latest/
    """
    # pylint: disable=import-outside-toplevel
    def __init__(self):
        # If import fails, the backend won't be used.
        from opentelemetry import trace
        self.otel_trace = trace

    def set_attribute(self, key, value):
        # Sets the value on the current span, not necessarily the root
        # span in the process.
        self.otel_trace.get_current_span().set_attribute(key, value)

    def record_exception(self):
        self.otel_trace.get_current_span().record_exception(sys.exc_info()[1])

    def create_span(self, name):
        # Currently, this is not implemented.
        pass

    def tag_root_span_with_error(self, exception):
        # Currently, this is not implemented for OTel
        pass

    def set_local_root_span_name(self, name, group=None, priority=None):
        # Currently this is not implemented
        pass


class DatadogBackend(TelemetryBackend):
    """
    Send telemetry to Datadog via ddtrace.

    API reference: https://ddtrace.readthedocs.io/en/stable/api.html
    """
    # pylint: disable=import-outside-toplevel
    def __init__(self):
        # If import fails, the backend won't be used.
        from ddtrace import tracer
        self.dd_tracer = tracer

    def set_attribute(self, key, value):
        if root_span := self.dd_tracer.current_root_span():
            root_span.set_tag(key, value)

    def record_exception(self):
        if span := self.dd_tracer.current_span():
            span.set_traceback()

    def create_span(self, name):
        return self.dd_tracer.trace(name)

    def tag_root_span_with_error(self, exception):
        root_span = self.dd_tracer.current_root_span()
        root_span.set_exc_info(type(exception), exception, exception.__traceback__)

    def set_local_root_span_name(self, name, group=None, priority=None):
        # For Datadog, this updates the 'resource_name' to the given name.
        local_root_span = self.dd_tracer.current_root_span()
        local_root_span.resource = name


# We're using an lru_cache instead of assigning the result to a variable on
# module load. With the default settings (pointing to a TelemetryBackend
# in this very module), this function can't be successfully called until
# the module finishes loading, otherwise we get a circular import error
# that will cause the backend to be dropped from the list.
@lru_cache
def configured_backends():
    """
    Produce a list of TelemetryBackend instances from Django settings.
    """
    # .. setting_name: OPENEDX_TELEMETRY
    # .. setting_default: ['edx_django_utils.monitoring.NewRelicBackend']
    # .. setting_description: List of telemetry backends to send data to. Allowable values
    #   are dotted module paths to classes implementing `edx_django_utils.monitoring.TelemetryBackend`,
    #   such as the built-in `NewRelicBackend`, `OpenTelemetryBackend`, and `DatadogBackend`
    #   (in the same module). For historical reasons, this defaults to just
    #   New Relic, and not all monitoring features will report to all backends (New Relic
    #   having the broadest support). Unusable options are ignored. Configuration
    #   of the backends themselves is via environment variables and system config files
    #   rather than via Django settings.
    backend_classes = getattr(settings, 'OPENEDX_TELEMETRY', None)
    if isinstance(backend_classes, str):
        # Prevent a certain kind of easy mistake.
        raise Exception("OPENEDX_TELEMETRY must be a list, not a string.")
    if backend_classes is None:
        backend_classes = ['edx_django_utils.monitoring.NewRelicBackend']

    backends = []
    for backend_class in backend_classes:
        try:
            cls = import_string(backend_class)
            if issubclass(cls, TelemetryBackend):
                backends.append(cls())
            else:
                log.warning(
                    f"Could not load OPENEDX_TELEMETRY option {backend_class!r}: "
                    f"{cls} is not a subclass of TelemetryBackend"
                )
        except BaseException as e:
            log.warning(f"Could not load OPENEDX_TELEMETRY option {backend_class!r}: {e!r}")

    return backends


@receiver(setting_changed)
def _reset_state(sender, **kwargs):  # pylint: disable=unused-argument
    """Reset caches when settings change during unit tests."""
    configured_backends.cache_clear()



"""
Methods to get plugin apps

Please remember to expose any new public methods in the `__init__.py` file.
"""
from logging import getLogger

from . import constants, registry

log = getLogger(__name__)


def get_plugin_apps(project_type):
    """
    Returns a list of all registered Plugin Apps, expected to be added to
    the INSTALLED_APPS list for the given project_type.
    """
    plugin_apps = [
        "{module_name}.{class_name}".format(
            module_name=app_config.__module__, class_name=app_config.__name__,
        )
        for app_config in registry.get_plugin_app_configs(project_type)
        if getattr(app_config, constants.PLUGIN_APP_CLASS_ATTRIBUTE_NAME, None)
        is not None
    ]
    # Note: This may not actually get logged; in edx-platform at least, logging seems to be
    # set at WARN at the time this is actually called.
    log.info(f"WARN: Plugin apps: For project type {project_type!r}, found {plugin_apps!r}")
    return plugin_apps



"""
Factories for models used in testing manufacture_data command
"""

import factory

from edx_django_utils.data_generation.tests.models import (
    TestCompany,
    TestContactInfo,
    TestNationalId,
    TestPerson,
    TestPersonContactPhoneNumber,
    test_model_nonstandard_casing
)


class TestNationalIdFactory(factory.django.DjangoModelFactory):
    """
    Test Factory for TestNationalId
    """

    class Meta:
        model = TestNationalId

    id_number = "123456789"


class TestPersonFactory(factory.django.DjangoModelFactory):
    """
    Test Factory for TestPerson
    """

    class Meta:
        model = TestPerson

    first_name = "John"
    last_name = "Doe"
    national_id = factory.SubFactory(TestNationalIdFactory)


class TestCompanyFactory(factory.django.DjangoModelFactory):
    """
    Test Factory for TestCompany
    """

    class Meta:
        model = TestCompany

    company_name = "Acme, Inc"
    national_id = factory.SubFactory(TestNationalIdFactory)
    ceo = factory.SubFactory(TestPersonFactory)


class TestContactInfoFactory(factory.django.DjangoModelFactory):
    """
    Test Factory for TestContactInfo
    """

    class Meta:
        model = TestContactInfo

    test_person = factory.SubFactory(TestPersonFactory)
    test_company = factory.SubFactory(TestCompanyFactory)
    address = "123 4th st, Fiveville, AZ, 67890"


class TestPersonContactPhoneNumberFactory(factory.django.DjangoModelFactory):
    """
    Test Factory for TestPersonContactPhoneNumber
    """

    class Meta:
        model = TestPersonContactPhoneNumber

    test_contact_info = factory.SubFactory(TestContactInfoFactory)
    phone_number = "(123) 456-7890"


class TestModelNonstandardCasingFactory(factory.django.DjangoModelFactory):
    """
    Test Factory for test_model_nonstandard_casing
    """

    class Meta:
        model = test_model_nonstandard_casing

    test_field = "TEST"


class AbstractFactory(factory.django.DjangoModelFactory):
    """
    Test factory for making sure our factory discovery doesn't choke on abstract factory classes
    """



"""
Tests for Data Generation
"""


from django.apps import AppConfig


class DataGenerationTestsConfig(AppConfig):
    name = 'edx_django_utils.data_generation.tests'
    label = 'data_generation_tests'  # Needed to avoid App label duplication with other tests modules



"""
Models used in testing manufacture_data command
"""
from django.db import models


class TestNationalId(models.Model):
    """
    For use in testing manufacture_data command
    """
    class Meta:
        app_label = 'data_generation_tests'

    id_number = models.CharField(max_length=10)


class TestPerson(models.Model):
    """
    For use in testing manufacture_data command
    """
    class Meta:
        app_label = 'data_generation_tests'

    first_name = models.CharField(max_length=30)
    last_name = models.CharField(max_length=30)
    national_id = models.ForeignKey(TestNationalId, null=True, on_delete=models.CASCADE)


class TestCompany(models.Model):
    """
    For use in testing manufacture_data command
    """
    class Meta:
        app_label = 'data_generation_tests'
    company_name = models.CharField(max_length=30)
    national_id = models.ForeignKey(TestNationalId, null=True, on_delete=models.CASCADE)
    ceo = models.ForeignKey(TestPerson, null=True, on_delete=models.CASCADE)


class TestContactInfo(models.Model):
    """
    For use in testing manufacture_data command
    """
    class Meta:
        app_label = 'data_generation_tests'
    test_person = models.ForeignKey(TestPerson, null=True, on_delete=models.CASCADE)
    test_company = models.ForeignKey(TestCompany, null=True, on_delete=models.CASCADE)
    address = models.CharField(max_length=100)


class TestPersonContactPhoneNumber(models.Model):
    """
    For use in testing manufacture_data command
    """
    class Meta:
        app_label = 'data_generation_tests'
    test_contact_info = models.ForeignKey(TestContactInfo, on_delete=models.CASCADE)
    phone_number = models.CharField(max_length=20)


class test_model_nonstandard_casing(models.Model):
    """
    For use in testing manufacture_data command
    """
    class Meta:
        app_label = 'data_generation_tests'
    test_field = models.CharField(max_length=30)



"""
Defines several custom monitoring helpers, some of which work with
the CachedCustomMonitoringMiddleware.

Usage:

    from edx_django_utils.monitoring import accumulate
    ...
    accumulate('xb_user_state.get_many.num_items', 4)

There is no need to do anything else.  The custom attributes are automatically
cleared before the next request.

We try to keep track of our custom monitoring at:
https://openedx.atlassian.net/wiki/spaces/PERF/pages/54362736/Custom+Attributes+in+New+Relic

At this time, the custom monitoring will only be reported to New Relic.

"""
from contextlib import ExitStack, contextmanager

from .backends import configured_backends
from .middleware import CachedCustomMonitoringMiddleware

try:
    import newrelic.agent
except ImportError:  # pragma: no cover
    newrelic = None  # pylint: disable=invalid-name


def accumulate(name, value):
    """
    Accumulate monitoring custom attribute for the current request.

    The named attribute is accumulated by a numerical amount using the sum.  All
    attributes are queued up in the request_cache for this request.  At the end of
    the request, the monitoring_utils middleware will batch report all
    queued accumulated attributes to the monitoring tool (e.g. New Relic).

    Arguments:
        name (str): The attribute name.  It should be period-delimited, and
            increase in specificity from left to right.  For example:
            'xb_user_state.get_many.num_items'.
        value (number):  The amount to accumulate into the named attribute.  When
            accumulate() is called multiple times for a given attribute name
            during a request, the sum of the values for each call is reported
            for that attribute.  For attributes which don't make sense to accumulate,
            use ``set_custom_attribute`` instead.

    """
    CachedCustomMonitoringMiddleware.accumulate_attribute(name, value)


def increment(name):
    """
    Increment a monitoring custom attribute representing a counter.

    Here we simply accumulate a new custom attribute with a value of 1, and the
    middleware should automatically aggregate this attribute.
    """
    accumulate(name, 1)


def set_custom_attributes_for_course_key(course_key):
    """
    Set monitoring custom attributes related to a course key.

    This is not cached.

    """
    set_custom_attribute('course_id', str(course_key))
    set_custom_attribute('org', str(course_key.org))


def set_custom_attribute(key, value):
    """
    Set monitoring custom attribute.

    This is not cached.
    """
    for backend in configured_backends():
        backend.set_attribute(key, value)


def record_exception():
    """
    Record a caught exception to the monitoring system.

    Note: By default, only unhandled exceptions are monitored. This function
    can be called to record exceptions as monitored errors, even if you handle
    the exception gracefully from a user perspective.
    """
    for backend in configured_backends():
        backend.record_exception()


@contextmanager
def function_trace(function_name):
    """
    Wraps a chunk of code that we want to appear as a separate, explicit,
    segment in our monitoring tools.
    """
    # Not covering this because if we mock it, we're not really testing anything
    # anyway. If something did break, it should show up in tests for apps that
    # use this code with whatever uses it.
    # ExitStack handles the underlying context managers.
    with ExitStack() as stack:
        for backend in configured_backends():
            context = backend.create_span(function_name)
            if context is not None:
                stack.enter_context(context)
        yield


def set_monitoring_transaction_name(name, group=None, priority=None):
    """
    Sets the name, group, and priority for the current root span.

    Current root span refers to the most ancestral span in the current process, rather than
    to the trace root, which may be outside of the process.

    Group and priority may not be supported by all backends.
    """
    for backend in configured_backends():
        backend.set_local_root_span_name(name, group=group, priority=priority)


def background_task(*args, **kwargs):
    """
    Handles monitoring for background tasks that are not passed in through the web server like
    celery and event consuming tasks.

    This function only supports New Relic.

    For more details, see:
    https://docs.newrelic.com/docs/apm/agents/python-agent/python-agent-api/backgroundtask-python-agent-api/

    """
    def noop_decorator(func):
        return func

    if newrelic:  # pragma: no cover
        return newrelic.agent.background_task(*args, **kwargs)
    else:
        return noop_decorator



"""
Monitoring utilities public api

Does not include signals.py, which is also part of the public api.
See README.rst for additional details.
"""
from .internal.backends import DatadogBackend, NewRelicBackend, OpenTelemetryBackend, TelemetryBackend
from .internal.code_owner.middleware import CodeOwnerMonitoringMiddleware
from .internal.code_owner.utils import (
    get_code_owner_from_module,
    set_code_owner_attribute,
    set_code_owner_attribute_from_module
)
from .internal.middleware import (
    CachedCustomMonitoringMiddleware,
    CookieMonitoringMiddleware,
    DeploymentMonitoringMiddleware,
    FrontendMonitoringMiddleware,
    MonitoringMemoryMiddleware,
    MonitoringSupportMiddleware
)
from .internal.transactions import ignore_transaction
from .internal.utils import (
    accumulate,
    function_trace,
    increment,
    record_exception,
    set_custom_attribute,
    set_custom_attributes_for_course_key,
    set_monitoring_transaction_name
)
# "set_custom_metric*" methods are deprecated
from .utils import set_custom_metric, set_custom_metrics_for_course_key



"""
Tests for MonitoringSupportMiddleware and associated utilities.

Note: See test_middleware.py for the rest of the middleware tests.
"""
from contextlib import contextmanager
from unittest.mock import ANY, Mock, call, patch

import ddt
from django.test import TestCase, override_settings

from edx_django_utils.cache import RequestCache
from edx_django_utils.monitoring import (
    CachedCustomMonitoringMiddleware,
    MonitoringSupportMiddleware,
    accumulate,
    increment
)
from edx_django_utils.monitoring.signals import (
    monitoring_support_process_exception,
    monitoring_support_process_request,
    monitoring_support_process_response
)

from ..utils import accumulate as deprecated_accumulate
from ..utils import increment as deprecated_increment
from ..utils import set_custom_attribute as deprecated_set_custom_attribute
from ..utils import set_custom_attributes_for_course_key as deprecated_set_custom_attributes_for_course_key


@ddt.ddt
class TestMonitoringSupportMiddleware(TestCase):
    """
    Test the monitoring_utils middleware and helpers
    """
    def setUp(self):
        super().setUp()
        self.mock_get_response = Mock()
        RequestCache.clear_all_namespaces()

    @patch('newrelic.agent')
    @ddt.data(
        (MonitoringSupportMiddleware, False, 'process_response'),
        (MonitoringSupportMiddleware, False, 'process_exception'),
        (CachedCustomMonitoringMiddleware, False, 'process_response'),
    )
    @ddt.unpack
    def test_accumulate_and_increment(
            self, cached_monitoring_middleware_class, is_deprecated, middleware_method_name, mock_newrelic_agent
    ):
        """
        Test normal usage of collecting custom attributes and reporting to New Relic
        """
        accumulate('hello', 10)
        accumulate('world', 10)
        accumulate('world', 10)
        increment('foo')
        increment('foo')

        # based on the attribute data above, we expect the following calls to newrelic:
        nr_agent_calls_expected = [
            call('hello', 10),
            call('world', 20),
            call('foo', 2),
        ]

        # fake a response to trigger attributes reporting
        middleware_method = getattr(cached_monitoring_middleware_class(self.mock_get_response), middleware_method_name)
        middleware_method(
            'fake request',
            'fake response',
        )

        # Assert call counts to newrelic.agent.add_custom_attribute()
        expected_call_count = len(nr_agent_calls_expected)
        if is_deprecated:
            expected_call_count += 1
        measured_call_count = mock_newrelic_agent.add_custom_attribute.call_count
        self.assertEqual(expected_call_count, measured_call_count)

        # Assert call args to newrelic.agent.add_custom_attribute().  Due to
        # the nature of python dicts, call order is undefined.
        mock_newrelic_agent.add_custom_attribute.assert_has_calls(nr_agent_calls_expected, any_order=True)

    @patch('newrelic.agent')
    @ddt.data(
        (MonitoringSupportMiddleware, False),
        (CachedCustomMonitoringMiddleware, False),
    )
    @ddt.unpack
    def test_accumulate_with_illegal_value(
            self, cached_monitoring_middleware_class, is_deprecated, mock_newrelic_agent
    ):
        """
        Test monitoring accumulate with illegal value that can't be added.
        """
        accumulate('hello', None)
        accumulate('hello', 10)

        # based on the metric data above, we expect the following calls to newrelic:
        nr_agent_calls_expected = [
            call('hello', None),
            call('error_adding_accumulated_metric', 'name=hello, new_value=10, cached_value=None'),
        ]

        self.mock_get_response = Mock()
        # fake a response to trigger metrics reporting
        cached_monitoring_middleware_class(self.mock_get_response).process_response(
            'fake request',
            'fake response',
        )

        # Assert call counts to newrelic.agent.add_custom_attribute()
        expected_call_count = len(nr_agent_calls_expected)
        if is_deprecated:
            expected_call_count += 1
        measured_call_count = mock_newrelic_agent.add_custom_attribute.call_count
        self.assertEqual(expected_call_count, measured_call_count)

        # Assert call args to newrelic.agent.add_custom_attribute().
        mock_newrelic_agent.add_custom_attribute.assert_has_calls(nr_agent_calls_expected, any_order=True)

    @contextmanager
    def catch_signal(self, signal):
        """
        Catch django signal and return the mocked handler.
        """
        handler = Mock()
        signal.connect(handler)
        yield handler
        signal.disconnect(handler)

    def test_process_request_signal(self):
        """
        Test middleware sends process request signal.
        """
        with self.catch_signal(monitoring_support_process_request) as handler:
            MonitoringSupportMiddleware(self.mock_get_response).process_request(
                'fake request'
            )

            handler.assert_called_once_with(
                signal=ANY, sender=MonitoringSupportMiddleware, request='fake request'
            )

    def test_process_response_signal(self):
        """
        Test middleware sends process response signal.
        """
        with self.catch_signal(monitoring_support_process_response) as handler:
            MonitoringSupportMiddleware(self.mock_get_response).process_response(
                'fake request', 'fake response'
            )

            handler.assert_called_once_with(
                signal=ANY, sender=MonitoringSupportMiddleware,
                request='fake request', response='fake response'
            )

    def test_process_exception_signal(self):
        """
        Test middleware sends process exception signal.
        """
        fake_exception = Exception()
        with self.catch_signal(monitoring_support_process_exception) as handler:
            MonitoringSupportMiddleware(self.mock_get_response).process_exception(
                'fake request', fake_exception
            )

            handler.assert_called_once_with(
                signal=ANY, sender=MonitoringSupportMiddleware,
                request='fake request', exception=fake_exception
            )

    @patch('ddtrace._trace.tracer.Tracer.current_root_span')
    def test_error_tagging(self, mock_get_root_span):
        # Ensure that we continue to support tagging exceptions in MonitoringSupportMiddleware.
        # This is only implemented for DatadogBackend at the moment.
        fake_exception = Exception()
        mock_root_span = Mock()
        mock_get_root_span.return_value = mock_root_span
        with override_settings(OPENEDX_TELEMETRY=['edx_django_utils.monitoring.DatadogBackend']):
            MonitoringSupportMiddleware(self.mock_get_response).process_exception(
                'fake request', fake_exception
            )
            mock_root_span.set_exc_info.assert_called_with(
                type(fake_exception), fake_exception, fake_exception.__traceback__
            )

    @patch('edx_django_utils.monitoring.utils.internal_accumulate')
    def test_deprecated_accumulate(self, mock_accumulate):
        deprecated_accumulate('key', 1)
        mock_accumulate.assert_called_with('key', 1)

    @patch('edx_django_utils.monitoring.utils.internal_increment')
    def test_deprecated_increment(self, mock_increment):
        deprecated_increment('key')
        mock_increment.assert_called_with('key')

    @patch('edx_django_utils.monitoring.utils.internal_set_custom_attribute')
    def test_deprecated_set_custom_attribute(self, mock_set_custom_attribute):
        deprecated_set_custom_attribute('key', True)
        mock_set_custom_attribute.assert_called_with('key', True)

    @patch('edx_django_utils.monitoring.utils.internal_set_custom_attributes_for_course_key')
    def test_deprecated_set_custom_attributes_for_course_key(self, mock_set_custom_attributes_for_course_key):
        deprecated_set_custom_attributes_for_course_key('key')
        mock_set_custom_attributes_for_course_key.assert_called_with('key')
